from mock import call, Mock, MagicMock, patch, PropertyMock

import pytest
import psycopg2
import psycopg2.errorcodes

from pg2kinesis.slot import SlotReader


@pytest.fixture
def slot():
    slot = SlotReader('blah_db', 'blah_host', 'blah_port', 'blah_user', 'blah_sslmode', 'pg2kinesis')
    slot._repl_cursor = Mock()
    slot._repl_conn = Mock()
    slot._normal_conn = Mock()

    return slot


def test__enter__(slot):
    # returns its self

    with patch.object(slot, '_get_connection', side_effects=[Mock(), Mock()]) as mock_gc:
        assert slot == slot.__enter__(), 'Returns itself'
        assert mock_gc.call_count == 2

        assert call.set_isolation_level(0) in slot._normal_conn.method_calls, 'make sure we are in autocommit'
        assert call.cursor() in slot._repl_conn.method_calls, 'we opened a cursor'

    with patch.object(slot, '_get_connection', side_effects=[Mock(), Mock()]) as mock_gc:
        slot.__enter__()


def test__exit__(slot):
    slot.__exit__(None, None, None)

    assert call.close() in slot._repl_cursor.method_calls
    assert call.close() in slot._repl_conn.method_calls
    assert call.close() in slot._normal_conn.method_calls

    slot._repl_cursor.close = Mock(side_effect=Exception)
    slot._repl_conn.close = Mock(side_effect=Exception)
    slot._normal_conn.close= Mock(side_effect=Exception)
    slot.__exit__(None, None, None)

    assert slot._repl_cursor.close.called, "Still called even thought call above raised"
    assert slot._repl_conn.close.called, "Still called even thought call above raised"
    assert slot._normal_conn.close.called, "Still called even thought call above raised"


def test_create_slot(slot):

    with patch.object(psycopg2.ProgrammingError, 'pgcode',
                      new_callable=PropertyMock,
                      return_value=psycopg2.errorcodes.DUPLICATE_OBJECT):
        pe = psycopg2.ProgrammingError()


        slot._repl_cursor.create_replication_slot = Mock(side_effect=pe)
        slot.create_slot()
        slot._repl_cursor.create_replication_slot.assert_called_with('pg2kinesis',
                                                                     slot_type=psycopg2.extras.REPLICATION_LOGICAL,
                                                                     output_plugin=u'test_decoding')
    with patch.object(psycopg2.ProgrammingError, 'pgcode',
                          new_callable=PropertyMock,
                          return_value=-1):
        pe = psycopg2.ProgrammingError()
        slot._repl_cursor.create_replication_slot = Mock(side_effect=pe)

        with pytest.raises(psycopg2.ProgrammingError) as e_info:
            slot.create_slot()
            slot._repl_cursor.create_replication_slot.assert_called_with('pg2kinesis',
                                                                         slot_type=psycopg2.extras.REPLICATION_LOGICAL,
                                                                         output_plugin=u'test_decoding')
        assert e_info.value.pgcode == -1

        slot._repl_cursor.create_replication_slot = Mock(side_effect=Exception)
    with pytest.raises(Exception):
        slot.create_slot()
        slot._repl_cursor.create_replication_slot.assert_called_with('pg2kinesis',
                                                                         slot_type=psycopg2.extras.REPLICATION_LOGICAL,
                                                                         output_plugin=u'test_decoding')


def test_delete_slot(slot):
    with patch.object(psycopg2.ProgrammingError, 'pgcode',
                      new_callable=PropertyMock,
                      return_value=psycopg2.errorcodes.UNDEFINED_OBJECT):
        pe = psycopg2.ProgrammingError()
        slot._repl_cursor.drop_replication_slot = Mock(side_effect=pe)
        slot.delete_slot()
    slot._repl_cursor.drop_replication_slot.assert_called_with('pg2kinesis')

    with patch.object(psycopg2.ProgrammingError, 'pgcode',
                      new_callable=PropertyMock,
                      return_value=-1):
        pe = psycopg2.ProgrammingError()
        slot._repl_cursor.create_replication_slot = Mock(side_effect=pe)
        with pytest.raises(psycopg2.ProgrammingError) as e_info:
            slot.delete_slot()
            slot._repl_cursor.drop_replication_slot.assert_called_with('pg2kinesis')

            assert e_info.value.pgcode == -1

    slot._repl_cursor.create_replication_slot = Mock(side_effect=Exception)
    with pytest.raises(Exception):
        slot.delete_slot()
        slot._repl_cursor.drop_replication_slot.assert_called_with('pg2kinesis')


def test__get_connection(slot):
    with patch('psycopg2.connect') as mock_connect:
        slot._get_connection()
        kall = call(connection_factory=None, cursor_factory=None, database='blah_db', host='blah_host',
                    port='blah_port', user='blah_user')
        assert mock_connect.called_with(kall)

        slot._get_connection(connection_factory='connection_fact', cursor_factory='cursor_fact')
        kall = call(connection_factory='connection_fact', cursor_factory='cursor_fact', database='blah_db', host='blah_host',
                    port='blah_port', user = 'blah_user')
        assert mock_connect.called_with(kall)


def test_primary_key_map(slot):
    slot._execute_and_fetch = Mock(return_value=[('test_table', 'pkey', 'uuid', 0),
                                                 ('test_table2', 'pkey', 'uuid', 0),
                                                 ('blue', 'bkey', 'char var', 10)
                                                 ])

    pkey_map = slot.primary_key_map

    assert len(pkey_map) == 3
    assert 'test_table' in pkey_map
    assert 'test_table2' in pkey_map
    assert 'blue' in pkey_map

    assert pkey_map['blue'].table_name == 'blue'
    assert pkey_map['blue'].col_name == 'bkey'
    assert pkey_map['blue'].col_type == 'char var'
    assert pkey_map['blue'].col_ord_pos == 10


def test_execute_and_fetch(slot):
    norm_conn = slot._normal_conn
    mock_cur = MagicMock()
    norm_conn.cursor = Mock(return_value=mock_cur)

    slot._execute_and_fetch('SQL SQL STATEMENT', 1, 2, 3)
    call.execute('SQL SQL STATEMENT', (1, 2, 3)) in mock_cur.method_calls
    assert call.__enter__().fetchall() in mock_cur.mock_calls

    mock_cur.reset_mock()
    slot._execute_and_fetch('SQL SQL STATEMENT')
    call.execute('SQL SQL STATEMENT') in mock_cur.method_calls
    assert call.__enter__().fetchall() in mock_cur.mock_calls


def test_process_replication_stream(slot):
    consume = Mock()
    slot.process_replication_stream(consume)

    assert call.start_replication('pg2kinesis', options=None) in  slot._repl_cursor.method_calls, 'We started replication event loop'
    assert call.consume_stream(consume) in slot._repl_cursor.method_calls, 'We pass consume to this method'




from __future__ import division
import time

import click

from .slot import SlotReader
from .formatter import get_formatter
from .stream import StreamWriter
from .log import logger

@click.command()
@click.option('--pg-dbname', '-d', help='Database to connect to.')
@click.option('--pg-host', '-h', default='',
              help='Postgres server location. Leave empty if localhost.')
@click.option('--pg-port', '-p', default='5432', help='Postgres port.')
@click.option('--pg-user', '-u', help='Postgres user')
@click.option('--pg-sslmode', help='Postgres SSL mode', default='prefer')
@click.option('--pg-slot-name', '-s', default='pg2kinesis',
              help='Postgres replication slot name.')
@click.option('--pg-slot-output-plugin', default='test_decoding',
              type=click.Choice(['test_decoding', 'wal2json']),
              help='Postgres replication slot output plugin')
@click.option('--stream-name', '-k', default='pg2kinesis',
              help='Kinesis stream name.')
@click.option('--message-formatter', '-f', default='CSVPayload',
              type=click.Choice(['CSVPayload', 'CSV']),
              help='Kinesis record formatter.')
@click.option('--table-pat', help='Optional regular expression for table names.')
@click.option('--full-change', default=False, is_flag=True,
              help='Emit all columns of a changed row.')
@click.option('--create-slot', default=False, is_flag=True,
              help='Attempt to on start create a the slot.')
@click.option('--recreate-slot', default=False, is_flag=True,
              help='Deletes the slot on start if it exists and then creates.')
def main(pg_dbname, pg_host, pg_port, pg_user, pg_sslmode, pg_slot_name, pg_slot_output_plugin,
         stream_name, message_formatter, table_pat, full_change, create_slot, recreate_slot):

    if full_change:
        assert message_formatter == 'CSVPayload', 'Full changes must be formatted as JSON.'
        assert pg_slot_output_plugin == 'wal2json', 'Full changes must use wal2json.'

    logger.info('Starting pg2kinesis')
    logger.info('Getting kinesis stream writer')
    writer = StreamWriter(stream_name)

    with SlotReader(pg_dbname, pg_host, pg_port, pg_user, pg_sslmode, pg_slot_name,
                    pg_slot_output_plugin) as reader:

        if recreate_slot:
            reader.delete_slot()
            reader.create_slot()
        elif create_slot:
            reader.create_slot()

        pk_map = reader.primary_key_map
        formatter = get_formatter(message_formatter, pk_map,
                                  pg_slot_output_plugin, full_change, table_pat)

        consume = Consume(formatter, writer)

        # Blocking. Responds to Control-C.
        reader.process_replication_stream(consume)

class Consume(object):
    def __init__(self, formatter, writer):
        self.cum_msg_count = 0
        self.cum_msg_size = 0
        self.msg_window_size = 0
        self.msg_window_count = 0
        self.cur_window = 0

        self.formatter = formatter
        self.writer = writer

    def __call__(self, change):
        self.cum_msg_count += 1
        self.cum_msg_size += change.data_size

        self.msg_window_size += change.data_size
        self.msg_window_count += 1

        fmt_msgs = self.formatter(change.payload)

        progress_msg = 'xid: {:12} win_count:{:>10} win_size:{:>10}mb cum_count:{:>10} cum_size:{:>10}mb'

        for fmt_msg in fmt_msgs:
            did_put = self.writer.put_message(fmt_msg)
            if did_put:
                change.cursor.send_feedback(flush_lsn=change.data_start)
                logger.info('Flushed LSN: {}'.format(change.data_start))

            int_time = int(time.time())
            if not int_time % 10 and int_time != self.cur_window:
                logger.info(progress_msg.format(
                    self.formatter.cur_xact, self.msg_window_count,
                    self.msg_window_size / 1048576, self.cum_msg_count,
                    self.cum_msg_size / 1048576))

                self.cur_window = int_time
                self.msg_window_size = 0
                self.msg_window_count = 0

if __name__ == '__main__':
    main()



import time
import aws_kinesis_agg.aggregator
import boto3

from botocore.exceptions import ClientError
from .log import logger

class StreamWriter(object):
    def __init__(self, stream_name, back_off_limit=60, send_window=13):
        self.stream_name = stream_name
        self.back_off_limit = back_off_limit
        self.last_send = 0

        self._kinesis = boto3.client('kinesis')
        self._sequence_number_for_ordering = '0'
        self._record_agg = aws_kinesis_agg.aggregator.RecordAggregator()
        self._send_window = send_window

        try:
            self._kinesis.create_stream(StreamName=stream_name, ShardCount=1)
        except ClientError as e:
            # ResourceInUseException is raised when the stream already exists
            if e.response['Error']['Code'] != 'ResourceInUseException':
                logger.error(e)
                raise

        waiter = self._kinesis.get_waiter('stream_exists')

        # waits up to 180 seconds for stream to exist
        waiter.wait(StreamName=self.stream_name)

    def put_message(self, fmt_msg):
        agg_record = None

        if fmt_msg:
            agg_record = self._record_agg.add_user_record(str(fmt_msg.change.xid), fmt_msg.fmt_msg)

        # agg_record will be a complete record if aggregation is full.
        if agg_record or (self._send_window and time.time() - self.last_send > self._send_window):
            agg_record = agg_record if agg_record else self._record_agg.clear_and_get()
            self._send_agg_record(agg_record)
            self.last_send = time.time()

        return agg_record

    def _send_agg_record(self, agg_record):
        if agg_record is None:
            return

        pk, _, data = agg_record.get_contents()
        logger.info('Sending %s records. Size %s. PK: %s' %
                    (agg_record.get_num_user_records(), agg_record.get_size_bytes(), pk))

        back_off = .05
        while back_off < self.back_off_limit:
            try:
                result = self._kinesis.put_record(Data=data,
                                                  PartitionKey=pk,
                                                  SequenceNumberForOrdering=self._sequence_number_for_ordering,
                                                  StreamName=self.stream_name)

            except ClientError as e:
                if e.response['Error']["Code"] == 'ProvisionedThroughputExceededException':
                    back_off *= 2
                    logger.warning('Provisioned throughput exceeded: sleeping %ss' % back_off)
                    time.sleep(back_off)
                else:
                    logger.error(e)
                    raise
            else:
                logger.debug('Sequence number: %s' % result['SequenceNumber'])
                break
        else:
            raise Exception('ProvisionedThroughputExceededException caused a backed off too many times!')



# coding=utf-8
from __future__ import unicode_literals
import json

import mock
import pytest

from pg2kinesis.slot import PrimaryKeyMapItem
from pg2kinesis.formatter import Change, CSVFormatter, CSVPayloadFormatter, Formatter, get_formatter


def get_formatter_produce_formatted_message(cls):
    # type: (Formatter, unicode) -> Message
    change = Change(xid=1, table=u'public.blue', operation=u'Update', pkey=u'123456')
    result = cls({}).produce_formatted_message(change)
    assert result.change == change
    return result


def test_CSVFormatter_produce_formatted_message():
    result = get_formatter_produce_formatted_message(CSVFormatter)

    assert result.fmt_msg == u'0,CDC,1,public.blue,Update,123456'


def test_CSVPayloadFormatter_produce_formatted_message():
    result = get_formatter_produce_formatted_message(CSVPayloadFormatter)
    assert result.fmt_msg.startswith(u'0,CDC,')
    payload = result.fmt_msg.split(',', 2)[-1]
    assert json.loads(payload) == dict(xid=1, table=u'public.blue', operation=u'Update', pkey=u'123456')


@pytest.fixture
def pkey_map():
    return {'public.test_table': PrimaryKeyMapItem(u'public.test_table', u'uuid', u'uuid', 0),
            'public.test_table2': PrimaryKeyMapItem(u'public.test_table2', u'name', u'character varying', 0)}


@pytest.fixture(params=[CSVFormatter, CSVPayloadFormatter, Formatter])
def formatter(request, pkey_map):
    return request.param(pkey_map)


def test___init__(formatter):
    # processes primary key map (adds colon)

    patterns = formatter._primary_key_patterns

    assert len(patterns) == 2
    assert u'public.test_table:' in patterns, 'with colon'
    assert u'public.test_table2:' in patterns, 'with colon'
    assert u'public.test_table' not in patterns, 'without colon should not be in patterns'
    assert u'public.test_table2' not in patterns, 'without colon should not be in patterns'
    assert patterns[u'public.test_table:'].pattern == r"uuid\[uuid\]:'?([\w\-]+)'?"
    assert patterns[u'public.test_table2:'].pattern == r"name\[character varying\]:'?([\w\-]+)'?"


def test__preprocess_test_decoding_change(formatter):
    # assert begin -> None + cur trans
    assert formatter.cur_xact == ''
    result = formatter._preprocess_test_decoding_change(u'BEGIN 100')
    assert result == []
    assert formatter.cur_xact == u'100'

    # assert commit -> None
    formatter.cur_xact = ''
    result = formatter._preprocess_test_decoding_change(u'COMMIT')
    assert result == []
    assert formatter.cur_xact == ''

    # error states
    with mock.patch.object(formatter, '_log_and_raise') as mock_log_and_raise:
        formatter._preprocess_test_decoding_change(u'UNKNOWN BLING')
        assert mock_log_and_raise.called
        mock_log_and_raise.assert_called_with(u'Unknown change: "UNKNOWN BLING"')

    with mock.patch.object(formatter, '_log_and_raise') as mock_log_and_raise:
        formatter._preprocess_test_decoding_change(u"table not_a_table: UPDATE: uuid[uuid]:'00079f3e-0479-4475-acff-4f225cc5188a'")
        assert mock_log_and_raise.called
        mock_log_and_raise.assert_called_with(u'Unable to locate table: "not_a_table:"')

    with mock.patch.object(formatter, '_log_and_raise') as mock_log_and_raise:
        formatter._preprocess_test_decoding_change(u"table public.test_table: UPDATE: not[not]:'00079f3e-0479-4475-acff-4f225cc5188a'")
        assert mock_log_and_raise.called
        mock_log_and_raise.assert_called_with(u'Unable to locate primary key for table "public.test_table"')

    # assert on proper match
    formatter.cur_xact = '1337'
    change = formatter._preprocess_test_decoding_change(
        u"table public.test_table: UPDATE: uuid[uuid]:'00079f3e-0479-4475-acff-4f225cc5188a'")[0]

    assert change.xid == u'1337'
    assert change.table == u'public.test_table'
    assert change.operation == u'UPDATE'
    assert change.pkey == u'00079f3e-0479-4475-acff-4f225cc5188a'

    change = formatter._preprocess_test_decoding_change(
        u"table public.test_table2: DELETE: name[character varying]:'Bling-2'")[0]

    assert change.xid == u'1337'
    assert change.table == u'public.test_table2'
    assert change.operation == u'DELETE'
    assert change.pkey == u'Bling-2'


def test__preprocess_wal2json_change(formatter):
    formatter.cur_xact = ''
    result = formatter._preprocess_wal2json_change(u"""{
                "xid": 101,
                "change": []
            }""")
    assert result == []
    assert formatter.cur_xact == 101

    with mock.patch.object(formatter, '_log_and_raise') as mock_log_and_raise:
        formatter._preprocess_wal2json_change(u"""{
                "xid": 100,
                "change": [
                    {
                        "kind": "insert",
                        "schema": "public",
                        "table": "not_a_table",
                        "columnnames": ["uuid"],
                        "columntypes": ["int4"],
                        "columnvalues": ["00079f3e-0479-4475-acff-4f225cc5188a"]
                    }
                ]
            }""")
        assert mock_log_and_raise.called
        mock_log_and_raise.assert_called_with(u'Unable to locate table: "public.not_a_table"')

    # assert on proper match
    formatter.cur_xact = '1337'
    change = formatter._preprocess_wal2json_change(u"""{
                "xid": 1337,
                "change": [
                    {
                        "kind": "insert",
                        "schema": "public",
                        "table": "test_table",
                        "columnnames": ["uuid"],
                        "columntypes": ["int4"],
                        "columnvalues": ["00079f3e-0479-4475-acff-4f225cc5188a"]
                    }
                ]
            }""")[0]

    assert change.xid == 1337
    assert change.table == u'public.test_table'
    assert change.operation == u'insert'
    assert change.pkey == u'00079f3e-0479-4475-acff-4f225cc5188a'

    change = formatter._preprocess_wal2json_change(u"""{
                "xid": 1337,
                "change": [
                    {
                        "kind": "delete",
                        "schema": "public",
                        "table": "test_table2",
                        "columnnames": ["name"],
                        "columntypes": ["varchar"],
                        "columnvalues": ["Bling-2"]
                    }
                ]
            }""")[0]

    assert change.xid == 1337
    assert change.table == u'public.test_table2'
    assert change.operation == u'delete'
    assert change.pkey == u'Bling-2'


def test__preprocess_wal2json_full_change(formatter):
    formatter.cur_xact = ''
    formatter.full_change = True

    result = formatter._preprocess_wal2json_change(u"""{
                "xid": 101,
                "change": []
            }""")
    assert result == []
    assert formatter.cur_xact == 101

    # Full changes from wal2json are not validated against known table map
    with mock.patch.object(formatter, '_log_and_raise') as mock_log_and_raise:
        formatter._preprocess_wal2json_change(u"""{
                "xid": 100,
                "change": [
                    {
                        "kind": "insert",
                        "schema": "public",
                        "table": "not_a_table",
                        "columnnames": ["uuid"],
                        "columntypes": ["int4"],
                        "columnvalues": ["00079f3e-0479-4475-acff-4f225cc5188a"]
                    }
                ]
            }""")
        assert not mock_log_and_raise.called

    # assert on proper match
    formatter.cur_xact = '1337'
    change = formatter._preprocess_wal2json_change(u"""{
                "xid": 1337,
                "change": [
                    {
                        "kind": "insert",
                        "schema": "public",
                        "table": "test_table",
                        "columnnames": ["uuid"],
                        "columntypes": ["int4"],
                        "columnvalues": ["00079f3e-0479-4475-acff-4f225cc5188a"]
                    }
                ]
            }""")[0]

    assert change.xid == 1337
    assert change.change == {
        "kind": "insert",
        "schema": "public",
        "table": "test_table",
        "columnnames": ["uuid"],
        "columntypes": ["int4"],
        "columnvalues": ["00079f3e-0479-4475-acff-4f225cc5188a"]
    }

    change = formatter._preprocess_wal2json_change(u"""{
                "xid": 1337,
                "change": [
                    {
                        "kind": "delete",
                        "schema": "public",
                        "table": "test_table2",
                        "columnnames": ["name"],
                        "columntypes": ["varchar"],
                        "columnvalues": ["Bling-2"]
                    }
                ]
            }""")[0]

    assert change.xid == 1337
    assert change.change == {
        "kind": "delete",
        "schema": "public",
        "table": "test_table2",
        "columnnames": ["name"],
        "columntypes": ["varchar"],
        "columnvalues": ["Bling-2"]
    }


def test_log_and_raise(formatter):

    with mock.patch('logging.Logger.error') as mock_log, pytest.raises(Exception) as e_info:
        formatter._log_and_raise(u'HELP!')

    assert str(e_info.value) == u'HELP!'
    mock_log.assert_called_with(u'HELP!')


def test___call__(formatter):
    with mock.patch.object(formatter, '_preprocess_test_decoding_change', return_value=[]) as mock_preprocess_test_decoding_change, \
            mock.patch.object(formatter, 'produce_formatted_message', return_value=None) as mock_pfm:
        assert not mock_preprocess_test_decoding_change.called
        result = formatter('COMMIT')
        assert mock_preprocess_test_decoding_change.called
        assert result == []
        assert not mock_pfm.called

    with mock.patch.object(formatter, '_preprocess_test_decoding_change', return_value=['blue']) as mock_preprocess_test_decoding_change, \
            mock.patch.object(formatter, 'produce_formatted_message', return_value='blue msg') as mock_pfm:
        assert not mock_preprocess_test_decoding_change.called
        result = formatter('blue message')
        assert mock_preprocess_test_decoding_change.called
        assert result == ['blue msg']
        mock_pfm.assert_called_with('blue')


def test_get_formatter():
    with mock.patch.object(Formatter, '__init__', return_value=None) as mocked:
        result = get_formatter('CSVPayload', 1, 2, 3, 4)
        assert isinstance(result, CSVPayloadFormatter)
        assert mocked.called
        mocked.assert_called_with(1, 2, 3, 4)

    with mock.patch.object(Formatter, '__init__', return_value=None) as mocked:
        result = get_formatter('CSV', 1, 2, 3, 4)
        assert isinstance(result, CSVFormatter)
        assert mocked.called
        mocked.assert_called_with(1, 2, 3, 4)



from __future__ import unicode_literals

import json
import re
import sys

from .log import logger

from collections import namedtuple

# Tuples representing changes as pulled from database
Change = namedtuple('Change', 'xid, table, operation, pkey')
FullChange = namedtuple('FullChange', 'xid, change')

# Final product of Formatter, a Change and the Change formatted.
Message = namedtuple('Message', 'change, fmt_msg')

COL_TYPE_VALUE_TEMPLATE_PAT = r"{col_name}\[{col_type}\]:'?([\w\-]+)'?"
MISSING_TABLE_ERR = 'Unable to locate table: "{}"'
MISSING_PK_ERR = 'Unable to locate primary key for table "{}"'

class Formatter(object):
    VERSION = 0
    TYPE = 'CDC'
    IGNORED_CHANGES = {'COMMIT'}

    def __init__(self, primary_key_map, output_plugin='test_decoding',
                 full_change=False, table_pat=None):

        self._primary_key_patterns = {}
        self.output_plugin = output_plugin
        self.primary_key_map = primary_key_map
        self.full_change = full_change
        self.table_pat = table_pat if table_pat is not None else r'[\w_\.]+'
        self.table_re = re.compile(self.table_pat)
        self.cur_xact = ''

        for k, v in getattr(primary_key_map, 'iteritems', primary_key_map.items)():
            # ":" added to make later look up not need to trim trailing ":".
            self._primary_key_patterns[k + ":"] = re.compile(
                COL_TYPE_VALUE_TEMPLATE_PAT.format(col_name=v.col_name, col_type=v.col_type)
            )

    def _preprocess_test_decoding_change(self, change):
        """
        Takes a message payload from the test_decoding plugin and distills it
        into a Change tuple currently only looking for primary key.

        They look like this:
            "table table_test: UPDATE: uuid[uuid]:'00079f3e-0479-4475-acff-4f225cc5188a' another_col[text]'bling'"

        :param change: a message payload from postgres' test_decoding plugin.
        :return: A list of type Change
        """

        rec = change.split(' ', 3)

        if rec[0] == 'BEGIN':
            self.cur_xact = rec[1]
        elif rec[0] in self.IGNORED_CHANGES:
            pass
        elif rec[0] == 'table':
            table_name = rec[1][:-1]

            if self.table_re.search(table_name):
                try:
                    mat = self._primary_key_patterns[rec[1]].search(rec[3])
                except KeyError:
                    self._log_and_raise(MISSING_TABLE_ERR.format(rec[1]))
                else:
                    if mat:
                        pkey = mat.groups()[0]
                        return [Change(xid=self.cur_xact, table=table_name,
                                       operation=rec[2][:-1], pkey=pkey)]
                    else:
                        self._log_and_raise(MISSING_PK_ERR.format(table_name))
        else:
            self._log_and_raise('Unknown change: "{}"'.format(change))

        return []

    def _preprocess_wal2json_change(self, change):
        """
        Takes a message payload from the wal2json plugin and distills it into a
        list of Change or FullChange tuples.

        They look like this:
            {
                "xid": 1234567890
                "change": [
                    {
                        "kind": "insert",
                        "schema": "public",
                        "table": "some_table",
                        "columnnames": ["id"],
                        "columntypes": ["int4"],
                        "columnvalues": [42]
                    }
                ]
            }
        :param change: a message payload from postgres wal2json plugin.
        :return: A list of type Change or FullChange
        """

        change_dictionary = json.loads(change)
        if not change_dictionary:
            return None

        self.cur_xact = change_dictionary['xid']
        changes = []

        for change in change_dictionary['change']:
            table_name = change['table']
            schema = change['schema']
            if self.table_re.search(table_name):
                if self.full_change:
                    changes.append(FullChange(xid=self.cur_xact, change=change))
                else:
                    try:
                        full_table = '{}.{}'.format(schema, table_name)
                        primary_key = self.primary_key_map[full_table]
                    except KeyError:
                        self._log_and_raise(MISSING_TABLE_ERR.format(full_table))
                    else:
                        value_index = change['columnnames'].index(primary_key.col_name)
                        pkey = str(change['columnvalues'][value_index])
                        changes.append(Change(xid=self.cur_xact,
                                              table=full_table,
                                              operation=change['kind'].lower(),
                                              pkey=pkey))
        return changes

    @staticmethod
    def _log_and_raise(msg):
        logger.error(msg)
        raise Exception(msg)

    def __call__(self, change):
        if self.output_plugin == 'test_decoding':
            pp_changes = self._preprocess_test_decoding_change(change)
        elif self.output_plugin == 'wal2json':
            pp_changes = self._preprocess_wal2json_change(change)
        return [self.produce_formatted_message(pp_change) for pp_change in pp_changes]

    def produce_formatted_message(self, change):
        return change


class CSVFormatter(Formatter):
    VERSION = 0
    def produce_formatted_message(self, change):
        fmt_msg = '{},{},{},{},{},{}'.format(CSVFormatter.VERSION,
                                             CSVFormatter.TYPE, *change)
        return Message(change=change, fmt_msg=fmt_msg)


class CSVPayloadFormatter(Formatter):
    VERSION = 0
    def produce_formatted_message(self, change):
        fmt_msg = '{},{},{}'.format(CSVFormatter.VERSION, CSVFormatter.TYPE,
                                    json.dumps(change._asdict()))
        return Message(change=change, fmt_msg=fmt_msg)


def get_formatter(name, primary_key_map, output_plugin, full_change, table_pat):
    formatter_f = getattr(sys.modules[__name__], '%sFormatter' % name)
    return formatter_f(primary_key_map, output_plugin, full_change, table_pat)



from __future__ import unicode_literals

from mock import Mock, call, patch

from pg2kinesis.__main__ import Consume

def test_consume():
    mock_formatter = Mock(return_value='fmt_msg')
    # required to avoid formatting error if cur_xact is logged
    mock_formatter.cur_xact = 'TEST_TRANSACTION'
    mock_writer = Mock()

    consume = Consume(mock_formatter, mock_writer)

    mock_change = Mock()
    mock_change.data_start = 10
    mock_change.data_size = 100
    mock_change.payload = 'PAYLOAD'

    mock_writer.put_message = Mock(return_value=False)
    consume(mock_change)
    assert mock_writer.put_message.called, 'Sanity'
    assert call.cursor.send_feedback(flush_lsn=10) not in mock_change.mock_calls, \
        'we did not send feedback!'

    mock_writer.put_message = Mock(return_value=True)
    consume(mock_change)
    assert mock_writer.put_message.called, 'Sanity'
    assert call.cursor.send_feedback(flush_lsn=10) in mock_change.mock_calls, \
        'we sent feedback!'


    mock_time = Mock()
    mock_time.return_value = 11.0

    consume.msg_window_size = 0
    consume.msg_window_count = 0
    consume.cur_window = 10
    with patch('time.time', mock_time):
        consume(mock_change)
        assert consume.cur_window == 10, 'cur window not updated if time is non-10-multiple'
        assert consume.msg_window_size == 100, 'msg_window_size not reset if time is non-10-multiple'
        assert consume.msg_window_count == 1, 'msg_window_count not reset if time is non-10-multiple'

    mock_time.return_value = 20.0
    with patch('time.time', mock_time):
        consume(mock_change)
        assert consume.cur_window == 20, 'cur window updated if time is multiple of 10'
        assert consume.msg_window_size == 0, 'msg_window_size reset if time is multiple of 10'
        assert consume.msg_window_count == 0, 'msg_window_count reset if time is multiple of 10'

    with patch('time.time', mock_time):
        consume(mock_change)
        assert consume.msg_window_size == 100, 'msg_window_size not reset if time is same as cur_window'
